/****************************************************************************
 *
 * bxcl3: BXCL mit Makros und Mengenklasse bxUnordColl
 *
 *  - Beispielprogramm zum Buch "Objektorientiertes Programmieren in C++",
 *    von Nicolai Josuttis, erschienen bei Addison-Wesley, ISBN 3-89319-637-4
 *
 *  - Abschnitt 5.1.7: Der Trick mit den Makros
 *    Abschnitt 5.2: Container- und Mengenklassen
 *  
 ****************************************************************************/

/****************************************************************************
 *
 * Copyright 1994 by Nicolai Josuttis and Addison-Wesley.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software for personal
 * and educational use is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice
 * and this permission notice appear in supporting documentation, and that
 * the names of Addison-Wesley or the authors are not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.  Addison-Wesley and the authors make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * ADDISON-WESLEY AND THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL ADDISON-WESLEY OR THE AUTHORS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 ****************************************************************************/

// da die meisten Compiler "bool" (noch) nicht koennen (siehe Seite 43):
typedef int bool;
#define true 1
#define false 0



/****************************************************************************
 *
 *	bxtypes.h
 *
 ****************************************************************************/

#ifndef _bxtypes_h
#define _bxtypes_h

/* Fuer Object-IDs
 *  - kann spaeter als Klasse deklariert werden
 */
typedef long bxid;
static const bxid bxnoid = -1;

#endif  /* _bxtypes_h */


/****************************************************************************
 *
 *	bxclass.h
 *
 ****************************************************************************/

#ifndef _bxclass_h
#define _bxclass_h

// Hilfstypen
//#include "bxtypes.h"

/* Hilfsklasse bxclass
 *  - beschreibt Eigenschaften jeder BXCL-Klasse
 *  - erlaubt eindeutige Klassenzuordnung
 *  - nicht zur Vererbung geeignet
 */
class bxclass {
  private:
    const char*    const className;      // Zeiger auf Klassenname
    const bxclass* const baseClassPtr;   // Zeiger auf Basisklasse

  public:
    // Konstruktor (mit Parametern fuer die konstanten Komponenten)
    bxclass (const char* const, const bxclass&);

    // Konstruktor speziell fuer bxObject (keine Basisklasse)
    bxclass (const char* const);

    // Gleichheit
    bool operator== (const bxclass&) const;
    bool operator!= (const bxclass& cl) const {
        return ! operator== (cl);
    }

    // Test auf Basisklasse
    bool isDerivedFrom (const bxclass&) const;

    // Klassen-Name
    const char* const name () const;
};

#endif  /* _bxclass_h */


/****************************************************************************
 *
 *	bxclass.cc / bxclass.cpp
 *
 ****************************************************************************/

/* Quelldatei zur Klasse bxclass
 */

// Header-Datei der eigenen Klasse einbinden
//#include "bxclass.h"

const bxclass* const NO_BASE_CLASS = ((bxclass*)0);

/* Konstruktor
 */
bxclass::bxclass (const char* name, const bxclass& base)
    : className(name), baseClassPtr(&base)
{
}

/* Konstruktor ohne Basisklasse
 */
bxclass::bxclass (const char* name)
    : className(name), baseClassPtr(NO_BASE_CLASS)
{
}
/* operator ==
 */
bool bxclass::operator== (const bxclass& cl) const
{
    return className == cl.className;
}

/* isDerivedFrom()
 *  - liefert, ob eine Klasse aus einer anderen Klasse
 *    abgeleitet, aber nicht identisch ist
 */
bool bxclass::isDerivedFrom (const bxclass& cl) const
{
    /* abgeleitet, wenn Basisklasse gleich oder abgeleitet
     */
    return baseClassPtr != NO_BASE_CLASS &&
           (*baseClassPtr == cl || baseClassPtr->isDerivedFrom(cl));
}

/* name()
 *  - liefert Klassenname
 */
const char* const bxclass::name () const
{
    return className;
}


/****************************************************************************
 *
 *	bxObject.h
 *
 ****************************************************************************/

#ifndef _bxObject_h
#define _bxObject_h

// Standard-Header-Dateien einbinden
#include <iostream.h>
#include <stdlib.h>

// Hilfstypen
//#include "bxtypes.h"
//#include "bxclass.h"

/* Abstrakte Klasse bxObject
 *  - Wurzel einer Klassenbibliothek
 *  - definiert alle Standard-Funktionen fuer Objekte
 */
class bxObject {
  /* Klassen-Eigenschaften
   */
  private:
    static const bxclass theClassDesc;
  public:
    static const bxclass& desc () {
        return theClassDesc;
    }
    virtual const bxclass& classDesc () const {
        return theClassDesc;
    }

    virtual bool isMemberOf (const bxclass& cl) const {
        return classDesc() == cl;
    }
    virtual bool isKindOf (const bxclass& cl) const {
        return classDesc() == cl
               || classDesc().isDerivedFrom(cl);
    }
    virtual const char* className () const {
        return classDesc().name();
    }

  /* Objekt-Eigenschaften
   */
  protected:
    static bxid maxId;    // aktuelle hoechste ID
    const  bxid objId;    // jeweilige Objekt-ID

  public:
    // Virtueller Destruktor
    virtual ~bxObject () {
    }

    /* Test auf identische Gleichheit
     *  - wird von allen Klassen einfach geerbt
     */
    virtual bool isSame (const bxObject&) const;

    /* Test auf Wertgleichheit mit beliebigem Objekt
     *  - basiert auf Test auf Wertgleichheit mit einem
     *    Objekt der gleichen Klasse (operator==)
     */
    virtual bool isEqual (const bxObject&) const = 0;

    // Ausgabe
    virtual void printOn (ostream& = cout) const = 0;

  protected:
    bxObject ()                            // Default-Konstruktor
      : objId(++maxId) {
    }
    bxObject (const bxObject&)             // Copy-Konstruktor
      : objId(++maxId) {
    }
    void operator = (const bxObject&) {    // Zuweisung
    }
};

/* Ausgabeoperator "<<" fuer alle Klassen deklarieren
 */
ostream& operator << (ostream& strm, const bxObject& obj);

/* Makro fuer Klassen-Deklaration (=> Header-Datei)
 *  - um gleichartige Deklarationen in allen Klassen abzunehmen
 */
#define BX_DECLARE_CLASS(CLASS,BASE,NAME)                     \
    private:                                                  \
      static const bxclass theClassDesc;                      \
    public:                                                   \
      static const bxclass& desc () {                         \
          return theClassDesc;                                \
      }                                                       \
      virtual const bxclass& classDesc () const {             \
          return theClassDesc;                                \
      }                                                       \
      virtual bool isEqual (const bxObject&) const;           \
      virtual bool operator!= (const CLASS& obj) const {      \
          return ! operator== (obj);                          \
      }

/* Makro fuer Klassen-Definitionen (=> Quelldatei)
 *  - um gleichartige Definitionen in allen Klassen abzunehmen
 */
#define BX_DEFINE_CLASS(CLASS,BASE,NAME)                      \
    const bxclass CLASS::theClassDesc (NAME, BASE::desc());   \
    bool CLASS::isEqual (const bxObject& obj) const           \
    {                                                         \
        return  classDesc() == obj.classDesc() &&             \
                *this == (const CLASS&)(obj);                 \
    }

#endif  /* _bxObject_h */


/****************************************************************************
 *
 *	bxObject.cc / bxObject.cpp
 *
 ****************************************************************************/

/* Quelldatei zur abstrakten Klasse bxObject
 */

// Header-Datei der eigenen Klasse einbinden
//#include "bxObject.h"

/* statische Komponente maxId initialisieren
 */
bxid bxObject::maxId = 0;

/* Klassen-Beschreibung
 *  - Name (keine Basisklasse)
 */
const bxclass bxObject::theClassDesc ("bxObject");

/* operator <<
 *  - Ausgabeoperator fuer alle Objekte
 */
ostream& operator << (ostream& strm, const bxObject& obj)
{
    // objektspezifische Ausgabe-Funktion aufrufen
    obj.printOn (strm);

    // fuer Verkettung
    return strm;
}

/* isSame
 *  - testet auf Identitaet durch Vergleich der IDs
 *  - geht auch bei Mehrfachvererbung
 */
bool bxObject::isSame (const bxObject& obj) const
{
    return objId == obj.objId;
}



/****************************************************************************
 *
 *	bxInt.h
 *
 ****************************************************************************/

#ifndef _bxInteger_h
#define _bxInteger_h

// Header-Datei der Basisklasse
//#include "bxObject.h"

/* Klasse bxInteger
 *  - abgeleitet von bxObject
 */
class bxInteger : public bxObject {

  /* Makro fuer die Deklarationen, die jede Klasse braucht und
   * im Prinzip gleich sind, aufrufen
   */
  BX_DECLARE_CLASS (bxInteger, bxObject, "bxInteger")

  protected:
    int theValue;                // Integer-Wert

  public:
    // Konstruktoren
    bxInteger (int i = 0) : theValue(i) {
    }

    // Wert abfragen oder aendern
    int value () const {
        return theValue;
    }
    int value (int i) {
        return theValue = i;
    }

    // Test auf Wertgleichheit mit Objekt gleicher Klasse
    virtual bool operator== (const bxInteger&) const;

    // Ausgabe
    virtual void printOn (ostream& = cout) const;

    // Copy-Konstruktor, da die Basisklasse ihn definiert
    bxInteger (const bxInteger& i)
        : bxObject(i), theValue(i.theValue) {
    }

    // Zuweisung, da die Basisklasse sie definiert
    const bxInteger& operator= (const bxInteger& i) {
        bxObject::operator=(i);    // Zuweisung der Basisklasse
        theValue = i.theValue;     // Wert zuweisen
        return *this;              // Objekt zurueckliefern
    }

    // virtueller Destruktor
    virtual ~bxInteger () {
    }
};

#endif  /* _bxInteger_h */


/****************************************************************************
 *
 *	bxInt.cc / bxInt.cpp
 *
 ****************************************************************************/

/* Quelldatei zur Klasse bxInteger
 */

// Header-Datei der eigenen Klasse einbinden
//#include "bxInteger.h"

/* Makro fuer die Definitionen, die jede Klasse braucht und
 * im Prinzip gleich sind, aufrufen
 */
BX_DEFINE_CLASS (bxInteger, bxObject, "bxInteger")

/* Test auf Wertgleichheit
 */
bool bxInteger::operator== (const bxInteger& x) const
{
    return  theValue == x.theValue;
}

/* Ausgabe
 */
void bxInteger::printOn (ostream& strm) const
{
    strm << theValue;
}


/****************************************************************************
 *
 *	bxString.h
 *
 ****************************************************************************/

#ifndef _bxString_h
#define _bxString_h

// Header-Datei der Basisklasse
//#include "bxObject.h"

/* Klasse bxString
 *  - abgeleitet von bxObject
 */
class bxString : public bxObject {

  /* Makro fuer die Deklarationen, die jede Klasse braucht und
   * im Prinzip gleich sind, aufrufen
   */
  BX_DECLARE_CLASS (bxString, bxObject, "bxString")

  protected:
    char*     cstring;    // Zeichenfolge als dynamischer C-String
    unsigned  len;        // Anzahl Zeichen (ohne '\0')
    unsigned  size;       // Speicherplatzgroesse von "cstring"

  public:
    // Default- und char*-Konstruktor
    bxString (const char* = "");

    // Aufgrund dynamischer Komponenten:
    bxString (const bxString&);                   // Copy-Konstruktor
    virtual ~bxString ();                         // virtueller Destruktor
    const bxString& operator= (const bxString&);  // Zuweisung

    // aneinanderhaengen von Strings
    virtual bxString operator+ (const bxString&) const;

    // Test auf Wertgleichheit mit Objekt gleicher Klasse
    virtual bool operator== (const bxString&) const;

    // Ausgabe
    virtual void printOn (ostream& = cout) const;

  protected:
    /* Konstruktor fuer uninitialisierten String bestimmter Laenge
     * (intern fuer operator+())
     */
    bxString (unsigned);
};

#endif  /* _bxString_h */


/****************************************************************************
 *
 *	bxString.cc / bxString.cpp
 *
 ****************************************************************************/

/* Quelldatei zur Klasse bxString
 */

// Header-Dateien einbinden
#include <string.h>
//#include "bxString.h"

/* Makro fuer die Definitionen, die jede Klasse braucht und
 * im Prinzip gleich sind, aufrufen
 */
BX_DEFINE_CLASS (bxString, bxObject, "bxString")

/* Default- und char*-Konstruktor
 *  - Default fuer s: ""
 */
bxString::bxString (const char* s)
{
    len = strlen (s);             // Anzahl Zeichen
    size = len + 1;               // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];    // Speicherplatz anlegen
    strcpy (cstring, s);          // Zeichen in Speicherplatz kopieren
}

/* Copy-Konstruktor
 */
bxString::bxString (const bxString& s)
    : bxObject (s)
{
    len = s.len;                    // Anzahl Zeichen uebernehmen
    size = len + 1;                 // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];      // Speicherplatz anlegen
    strcpy (cstring, s.cstring);    // Zeichen kopieren
}

/* Destruktor
 */
bxString::~bxString ()
{
    // Mit new angelegten Speicherplatz wieder freigeben
    delete [] cstring;
}

/* Zuweisung (Operator =)
 */
const bxString& bxString::operator= (const bxString& s)
{
    // Zuweisung eines Strings an sich selbst macht nichts
    if (isSame(s)) {
        return *this;        // String zurueckliefern    
    }

    // Zuweisungen der Basisklasse durchfuehren
    bxObject::operator= (s);

    len = s.len;           // Anzahl Zeichen uebernehmen

    // falls Platz nicht reicht, vergroessern
    if (size < len + 1) {
        delete [] cstring;            // alten Speicherplatz freigeben
        size = len + 1;               // Anzahl Zeichen plus '\0'
        cstring = new char [size];    // Speicherplatz anlegen
    }

    strcpy (cstring, s.cstring);      // Zeichen kopieren

    return *this;            // geaenderten String zurueckliefern    
}

/* Hintereinanderhaengen (Operator +)
 */
bxString bxString::operator+ (const bxString& s) const
{
    // uninitialisierten Summen-String erzeugen
    bxString sum (len + s.len + 1);
 
    // Summen-Zeichenfolge darin initialisieren
    strcpy (sum.cstring, cstring);
    strcat (sum.cstring, s.cstring);

    // Kopie von Summen-String zurueck
    return sum;
}

/* Konstruktor fuer uninitialisierten String bestimmter Laenge
 * (intern fuer operator+())
 */
bxString::bxString (unsigned l)
{
    len = l;                      // Anzahl Zeichen uebernehmen
    size = len + 1;               // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];    // Speicherplatz anlegen
}

bool bxString::operator== (const bxString& x) const
{
    return  (strcmp(cstring,x.cstring) == 0);
}

void bxString::printOn (ostream& strm) const
{
    // Zeichenfolge einfach ausgeben
    strm << cstring;
}


/****************************************************************************
 *
 *	bxUnordColl.h
 *
 ****************************************************************************/

#ifndef _bxUnordColl_h
#define _bxUnordColl_h

// Header-Datei der Basisklasse
//#include "bxObject.h"

/* Klasse bxUnordColl
 *  - abgeleitet von bxObject
 */
class bxUnordColl : public bxObject {

  /* Makro fuer die Deklarationen, die jede Klasse braucht und
   * im Prinzip gleich sind, aufrufen
   */
  BX_DECLARE_CLASS (bxUnordColl, bxObject, "bxUnordColl")

  protected:
    bxObject** elems;      // dynamisches Array von bxObject-Zeigern
    unsigned   size;       // Array-Groesse
    unsigned   num;        // Anzahl eingetragener Elemente

  public:
    // Default-Konstruktor
    bxUnordColl ();

    // Funktionen fuer dynamische Komponenten
    bxUnordColl (const bxUnordColl&);
    virtual ~bxUnordColl();
    const bxUnordColl& operator= (const bxUnordColl&);

    // Element dazufuegen / entfernen
    virtual void add (bxObject&);
    virtual void remove (const bxObject&);

    // Array-Groesse aendern
    virtual void resize (unsigned int);

    // Anzahl Elemente liefern
    virtual unsigned numElems () const {
        return num;
    }

    // Standard-Funktionen
    virtual bool operator== (const bxUnordColl&) const;
    virtual void printOn (ostream& strm) const;
};

#endif



/****************************************************************************
 *
 *	bxUnordColl.cc / bxUnordColl.cpp
 *
 ****************************************************************************/

/* Quelldatei zur Klasse bxUnordColl
 */

// Header-Datei der eigenen Klasse einbinden
//#include "bxUnordColl.h"

/* Makro fuer die Definitionen, die jede Klasse braucht und
 * im Prinzip gleich sind, aufrufen
 */
BX_DEFINE_CLASS (bxUnordColl, bxObject, "bxUnordColl")

/* Konstruktor
 */
bxUnordColl::bxUnordColl ()
{
    size = 0;                        // Platzbedarf ist 0
    elems = new bxObject* [size];    // (leeres) Array anlegen
    num = 0;                         // kein Element drin
}

/* Copy-Konstruktor
 */
bxUnordColl::bxUnordColl (const bxUnordColl& m)
    : bxObject(m)                    // Copy-Konstruktor der Basisklasse
{
    num = m.num;                     // Anzahl der Elemente uebernehmen
    size = num;                      // das ist auch der Platzbedarf
    elems = new bxObject* [size];    // Array anlegen

    // Elemente (Objekt-Zeiger) kopieren
    for (unsigned i=0; i<num; i++) {
        elems[i] = m.elems[i];
    }
}

/* Destruktor
 */
bxUnordColl::~bxUnordColl ()
{
    delete [] elems;                 // Array freigeben        
}

/* Zuweisung =
 */
const bxUnordColl& bxUnordColl::operator= (const bxUnordColl& m)
{
    // Zuweisung an sich selbst macht nichts
    if (isSame(m)) {      
        return *this;
    }

    // Zuweisungen der Basisklasse durchfuehren
    bxObject::operator= (m);

    num = m.num;                     // Anzahl der Elemente uebernehmen
    
    // Falls zu wenig Platz, vergroessern
    if (num > size) {
        delete [] elems;
        size = num;
        elems = new bxObject* [size];
    }

    // einzelne Elemente zuweisen
    for (unsigned i=0; i<num; i++) {
        elems[i] = m.elems[i];
    }

    return *this;
}

/* add
 *  - Element eintragen (ggfs. vergroessern)
 */
void bxUnordColl::add (bxObject& obj)
{
    if (num >= size) {
        // Groesse bei 0 auf 10 setzen oder sonst verdoppeln
        resize (size==0 ? 10 : 2*size);
    }

    // Adresse von Objekt eintragen und Elemente-Anzahl erhoehen
    elems[num++] = &obj;
}

/* remove
 *  - erstes identisches Element entfernen
 */
void bxUnordColl::remove (const bxObject& obj)
{
    for (unsigned i=0; i<num; i++) {
        if (elems[i]->isSame(obj)) {
            /* identisches Objekt gefunden
             *  - letztes Element (Objekt-Zeiger) dort eintragen
             *  - und tschuess (keine weiteren suchen)
             */
            elems[i] = elems[--num];
            return;
        }
    }
}

/* resize
 */
void bxUnordColl::resize (unsigned int newSize)
{
    // newSize zu klein ?
    if (newSize < num) {
        newSize = num;
    }

    // neues Array mit neuer Groesse anlegen
    size = newSize;
    bxObject** tmp = new bxObject* [size];

    // bisherige Objekte kopieren
    for (unsigned i=0; i<num; i++) {
        tmp[i] = elems[i];
    }

    // altes Array freigeben und neues verwenden
    delete [] elems;
    elems = tmp;
}

/* printOn
 */
void bxUnordColl::printOn (ostream& strm) const
{
    /* Elemente umgeben von runden Klammern und durch Leerzeichen
     * getrennt ausgeben
     */
    strm << "( ";
    for (unsigned i=0; i<num; i++) {
        strm << *elems[i] << ' ';
    }
    strm << ')';
}


/* operator ==
 */
bool bxUnordColl::operator== (const bxUnordColl& m) const
{
    /* Trivialfaelle
     *  - ungleiche Anzahl
     *  - beide kein Element
     *  - beide ein Element
     */
    if (num != m.num) {
        return false;
    }
    if (num == 0) {
        return true;
    }
    if (num == 1) {
        return elems[0] == m.elems[0];
    }

    /* Da in einer ungeordneten Menge keine Reihenfolge definiert
     * ist, muss festgestellt werden, ob es zu jedem Element in der
     * einen Menge genau ein Element in der anderen Menge gibt.
     */
    bool* abgehakt = new bool [num];
    unsigned i, j;

    // Zunaechst keines gefunden
    for (i = 0; i < num; i++) {
        abgehakt[i] = false;
    }

    /* Jedes Element in m zu finden (und abzuhaken) ?
     */
    for (i = 0; i < num; i++) {
         bool gefunden = false;
         for (j = 0; j < num && !gefunden; j++) {
             if (!abgehakt[j] && elems[i]->isSame(*m.elems[j])) {
                 gefunden = true;
                 abgehakt[j] = true;
             }
         }
         if (!gefunden) {
             // ein Element ist nicht in der andere Menge
             delete [] abgehakt;
             return false;
         }
    }

    // Alle Elemente sind genau einmal in der anderen Mange
    delete [] abgehakt;
    return true;
}


/****************************************************************************
 *
 *	bxcl3.cc / bxcl3.cpp
 *
 ****************************************************************************/

// Header-Dateien
#include <iostream.h>
//#include "bxInteger.h"
//#include "bxString.h"
//#include "bxUnordColl.h"

main()
{
    bxString s1, s2, s3;
    bxInteger i1;
    bxUnordColl c1, c2;

    s1 = s2 = "hallo";
    s3 = "*****";
    i1 = 17;

    c1.add (s1);        // Menge c1 enthaelt: - bxString s1
    c1.add (s2);        //                    - bxString s2
    c1.add (s3);        //                    - bxString s3
    c1.add (i1);        //                    - bxInteger i1
    c1.add (s2);        //                    - bxString s2
    c1.add (s3);        //                    - bxString s3

    c2 = c1;            // Menge c1 wird an Menge c2 zugeweisen
    c2.remove (s2);     // Menge c2 verliert: - bxString s2
    c2.add (s2);        // Menge c2 bekommt:  - bxString s2

    cout << "c1: " << c1 << endl;
    cout << "c2: " << c2 << endl;
    if (c1 == c2) {
        cout << "c1 == c2" << endl;
    }
    else {
        cout << "c1 != c2" << endl;
    }
}


